/*
This file is part of OCamlRPM.

OCamlRPM is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

OCamlRPM is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OCamlRPM; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
 
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include <rpmlib.h>

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>

#undef DEBUG

/* pkg_priority */
#define PRIO_REQUIRED 0
#define PRIO_IMPORTANT 1
#define PRIO_STANDARD 2
#define PRIO_OPTIONAL 3
#define PRIO_EXTRA 4

/* pkg_metadata fields */
#define METADATA_ARCHITECTURE 0 /* RPMTAG_ARCH */
#define METADATA_BUGS 1
#define METADATA_BUILD_ESSENTIAL 2 /* ignore */
#define METADATA_CONFLICTS 3 /* RPMTAGS_CONFLICTFLAGS, RPMTAGS_CONFLICTNAME, RPMTAGS_CONFLICTVERSUIN */
#define METADATA_DEPENDS 4 /* RPMTAGS_REQUIREFLAGS, RPMTAGS_REQUIRENAME, RPMTAGS_REQUIREVERSION */
#define METADATA_DESCRIPTION 5 /* RPMTAG_DESCRIPTION */
#define METADATA_DISTRIBUTION 6 /* RPMTAG_DISTRIBUTION */
#define METADATA_ENHANCES 7 /* ignore */
#define METADATA_ESSENTIAL 8 /* ignore */
#define METADATA_FILENAME 9
#define METADATA_INSTALLED_SIZE 10 /* RPMTAG_SIZE */
#define METADATA_LICENSE 11 /* RPMTAG_LICENSE */
#define METADATA_MAINTAINER 12 /* RPMTAG_PACKAGER */
#define METADATA_MD5SUM 13
#define METADATA_NAME 14 /* RPMTAG_NAME */
#define METADATA_ORIGIN 15
#define METADATA_OS 16 /* RPMTAG_OS */
#define METADATA_PRE_DEPENDS 17 /* ignore */
#define METADATA_PRIORITY 18
#define METADATA_PROVIDES 19 /* RPMTAG_PROVIDENAME, RPMTAG_PROVIDEFLAGS, RPMTAG_PROVIDEVERSION */
#define METADATA_RECOMMENDS 20 /* ignore */
#define METADATA_REPLACES 21 /* RPMTAG_OBSOLETENAME, RPMTAG_OBSOLETEFLAGS, RPMTAG_OBSOLETEVERSION */
#define METADATA_SECTION 22 /* RPMTAG_GROUP */
#define METADATA_SHA1SUM 23
#define METADATA_SHA256SUM 24
#define METADATA_SIZE 25 /* RPMTAG_ARCHIVESIZE */
#define METADATA_SOURCE_PACKAGE 26 /* RPMTAG_SOURCERPM */
#define METADATA_SOURCE_VERSION 27 /* RPMTAG_SOURCERPM */
#define METADATA_SUGGESTS 28 /* ignore */
#define METADATA_SUMMARY 29 /* RPMTAG_SUMMARY */
#define METADATA_TAGS 30
#define METADATA_TASKS 31
#define METADATA_URL 32 /* RPMTAG_URL */
#define METADATA_VENDOR 33 /* RPMTAG_VENDOR */
#define METADATA_VERSION 34 /* RPMTAG_VERSION, RPMTAG_RELEASE, RPMTAG_EPOCH */
#define METADATA_UNKNOWN 35
#define METADATA_FILES 36 /* RPMTAG_DIRINDEXES, RPMTAG_DIRNAMES, RPMTAG_BASENAMES */
#define METADATA_NR_FIELDS 37

char *ocamlize_string (int_32 my_locale, int_32 nr_locales, int_32 type, char *data)
{
	int_32 i;
	char *ptr;
	switch (type)
	{
		case RPM_STRING_TYPE:
			return data;
		case RPM_I18NSTRING_TYPE:
			for (i = 0, ptr = data + (nr_locales * 4); i < my_locale; i++, ptr += strlen (ptr) + 1);
#ifdef DEBUG
			fprintf (stderr, "I18N string (locale %d): %s\n", my_locale, ptr);
#endif
			return ptr;
		default:
			fprintf (stderr, "ocamlize_string called for something that is not a string.\n");
			exit (EXIT_FAILURE);
	}
}

/* Returns a good version specification */
value ocamlize_dependency (char *name, char *version, int_32 flags)
{
	CAMLparam0 ();
	CAMLlocal2 (dep, v0);
	dep = caml_alloc (2, 0);
	Store_field (dep, 0, caml_copy_string (name));
	switch (flags % 16)
	{
		case 2: /* < x */
			v0 = caml_alloc (1, 0);
			Store_field (v0, 0, caml_copy_string (version));
			Store_field (dep, 1, v0);
			break;
		case 4: /* > x */
			v0 = caml_alloc (1, 4);
			Store_field (v0, 0, caml_copy_string (version));
			Store_field (dep, 1, v0);
			break;
		case 8: /* = x */
			v0 = caml_alloc (1, 2);
			Store_field (v0, 0, caml_copy_string (version));
			Store_field (dep, 1, v0);
			break;
		case 10: /* <= x */
			v0 = caml_alloc (1, 1);
			Store_field (v0, 0, caml_copy_string (version));
			Store_field (dep, 1, v0);
			break;
		case 12: /* >= x */
			v0 = caml_alloc (1, 3);
			Store_field (v0, 0, caml_copy_string (version));
			Store_field (dep, 1, v0);
			break;
		default: /* any version */
			Store_field (dep, 1, Val_int (0)); /* AnyVersion */
	}
	CAMLreturn (dep);
}

/* Returns a pkg_metadata value */
value ocamlize_header (Header header, char *locale)
{
	HeaderIterator hi;
	int_32 tag, type, count, epoch = 0, *dirindexes = NULL, nr_di = 0, nr_bn = 0, nr_dn = 0, nr_md5s = 0, nr_p = 0, nr_d = 0, nr_c = 0, nr_o = 0, i, *p_flags = NULL, *d_flags = NULL, *c_flags = NULL, *o_flags = NULL, nr_locales = 0, my_locale = 0;
	uint_32 installed_size, archive_size, date = 0;
	char *version = NULL, *release = NULL, **basenames = NULL, **dirnames = NULL, *version_string, **p_names = NULL, **p_versions = NULL, **d_names = NULL, **d_versions = NULL, **c_names = NULL, **c_versions = NULL, **o_names = NULL, **o_versions = NULL, **locales = NULL, *name = NULL, **md5s = NULL, *arch = NULL, *maintainer = NULL, *section = NULL, *source_package = NULL, *url = NULL, *summary = NULL, *description = NULL, *license = NULL, *distribution = NULL, *vendor = NULL, *os = NULL, *filename = NULL; 
	const void *data;
	CAMLparam0 ();
	CAMLlocal5 (metadata, v0, v1, v2, v3);
	CAMLlocal4 (v4, v5, v6, v7);
	metadata = caml_alloc (METADATA_NR_FIELDS, 0);
	hi = headerInitIterator (header);
	while (headerNextIterator (hi, &tag, &type, &data, &count))
	{
		switch (tag)
		{
			case HEADER_I18NTABLE:
				locales = (char **) malloc (count * sizeof (char *));
				for (i = 0; i < count; i++)
				{
					locales [i] = strdup (((char **) data) [i]);
					if (strcmp (locales [i], locale) == 0)
						my_locale = i;
				}
				nr_locales = count;
				break;
			case RPMTAG_NAME: /* string type */
#ifdef DEBUG
				fprintf (stderr, "- Name: %s\n", (char *) data);
#endif
				if (type == RPM_STRING_TYPE)
					name = strdup ((char *) data);
				else
					caml_failwith ("NAME tag without string type");	
				break;
			case RPMTAG_EPOCH:
				if (type == RPM_INT32_TYPE)
					epoch = ((int_32 *) data) [0];
				else
					caml_failwith ("EPOCH tag without int32 type");
				break;
			case RPMTAG_VERSION:
				if (type == RPM_STRING_TYPE)
					version = strdup ((char *) data);
				else
					caml_failwith ("VERSION tag without string type");
				break;
			case RPMTAG_RELEASE:
				if (type == RPM_STRING_TYPE)
					release = strdup ((char *) data);
				else
					caml_failwith ("RELEASE tag without string type");
				break;
			case RPMTAG_REQUIRENAME:
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					nr_d = count;
					d_names = (char **) malloc (nr_d * sizeof (char *));
					for (i = 0; i < nr_d; i++)
						d_names [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("REQUIRENAME tag without string type");
				break;
			case RPMTAG_REQUIREVERSION:
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					if (nr_d == 0)
						nr_d = count;	
					else if (count != nr_d)
						caml_failwith ("The number of REQUIREVERSIONS is not consistent");
					d_versions = (char **) malloc (nr_d * sizeof (char *));
					for (i = 0; i < nr_d; i++)
						d_versions [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("REQUIREVERSION tag without string type");
				break;
			case RPMTAG_REQUIREFLAGS:
				if (type == RPM_INT32_TYPE)
				{
					if (nr_d == 0)
						nr_d = count;	
					else if (count != nr_d)
						caml_failwith ("The number of REQUIREFLAGS is not consistent");
					d_flags = (int_32 *) malloc (nr_d * sizeof (int_32));
					for (i = 0; i < nr_d; i++)
						d_flags [i] = ((int_32 *) data) [i];
				}
				else
					caml_failwith ("REQUIREFLAGS tag without string type");
				break;
			case RPMTAG_PROVIDENAME:
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					nr_p = count;
					p_names = (char **) malloc (nr_p * sizeof (char *));
					for (i = 0; i < nr_p; i++)
						p_names [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("PROVIDENAME tag without string type");
				break;
			case RPMTAG_PROVIDEVERSION:
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					if (nr_p == 0)
						nr_p = count;	
					else if (count != nr_p)
						caml_failwith ("The number of PROVIDEVERSIONS is not consistent");
					p_versions = (char **) malloc (nr_p * sizeof (char *));
					for (i = 0; i < nr_p; i++)
						p_versions [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("PROVIDEVERSION tag without string type");
				break;
			case RPMTAG_PROVIDEFLAGS:
				if (type == RPM_INT32_TYPE)
				{
					if (nr_p == 0)
						nr_p = count;	
					else if (count != nr_p)
						caml_failwith ("The number of PROVIDEFLAGS is not consistent");
					p_flags = (int_32 *) malloc (nr_p * sizeof (int_32));
					for (i = 0; i < nr_p; i++)
						p_flags [i] = ((int_32 *) data) [i];
				}
				else
					caml_failwith ("PROVIDEFLAGS tag without string type");
				break;
			case RPMTAG_CONFLICTNAME:
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					nr_c = count;
					c_names = (char **) malloc (nr_c * sizeof (char *));
					for (i = 0; i < nr_c; i++)
						c_names [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("CONFLICTNAME tag without string type");
				break;
			case RPMTAG_CONFLICTVERSION:
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					if (nr_c == 0)
						nr_c = count;	
					else if (count != nr_c)
						caml_failwith ("The number of CONFLICTVERSIONS is not consistent");
					c_versions = (char **) malloc (nr_c * sizeof (char *));
					for (i = 0; i < nr_c; i++)
						c_versions [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("CONFLICTVERSION tag without string type");
				break;
			case RPMTAG_CONFLICTFLAGS:
				if (type == RPM_INT32_TYPE)
				{
					if (nr_c == 0)
						nr_c = count;	
					else if (count != nr_c)
						caml_failwith ("The number of CONFLICTFLAGS is not consistent");
					c_flags = (int_32 *) malloc (nr_c * sizeof (int_32));
					for (i = 0; i < nr_c; i++)
						c_flags [i] = ((int_32 *) data) [i];
				}
				else
					caml_failwith ("CONFLICTFLAGS tag without string type");
				break;
			case RPMTAG_OBSOLETENAME:
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					nr_o = count;
					o_names = (char **) malloc (nr_o * sizeof (char *));
					for (i = 0; i < nr_o; i++)
						o_names [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("OBSOLETENAME tag without string type");
				break;
			case RPMTAG_OBSOLETEVERSION:
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					if (nr_o == 0)
						nr_o = count;	
					else if (count != nr_o)
						caml_failwith ("The number of OBSOLETEVERSIONS is not consistent");
					o_versions = (char **) malloc (nr_o * sizeof (char *));
					for (i = 0; i < nr_o; i++)
						o_versions [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("OBSOLETEVERSION tag without string type");
				break;
			case RPMTAG_OBSOLETEFLAGS:
				if (type == RPM_INT32_TYPE)
				{
					if (nr_o == 0)
						nr_o = count;	
					else if (count != nr_o)
						caml_failwith ("The number of OBSOLETEFLAGS is not consistent");
					o_flags = (int_32 *) malloc (nr_o * sizeof (int_32));
					for (i = 0; i < nr_o; i++)
						o_flags [i] = ((int_32 *) data) [i];
				}
				else
					caml_failwith ("OBSOLETEFLAGS tag without string type");
				break;
			case RPMTAG_ARCH: /* string type */
				if (type == RPM_STRING_TYPE)
					arch = strdup ((char *) data);
				else
					caml_failwith ("ARCHITECTURE tag without string type");	
				break;
			case RPMTAG_DIRINDEXES:
				nr_di = count;
				if (type == RPM_INT32_TYPE)
				{
					dirindexes = (int_32 *) malloc (nr_di * sizeof (int_32 *));
					for (i = 0; i < nr_di; i++)
						dirindexes [i] = ((int_32 *) data) [i];
				}
				else
					caml_failwith ("DIRINDEXES tag without int32 type");
				break;
			case RPMTAG_BASENAMES:
				nr_bn = count;
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					basenames = (char **) malloc (nr_bn * sizeof (char *));
					for (i = 0; i < nr_bn; i++)
						basenames [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("BASENAMES tag without string array type");
				break;
			case RPMTAG_DIRNAMES:
				nr_dn = count;
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					dirnames = (char **) malloc (nr_dn * sizeof (char *));
					for (i = 0; i < nr_dn; i++)
						dirnames [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("BASENAMES tag without string array type");
				break;
			case RPMTAG_FILEMD5S:
				nr_md5s = count;
				if (type == RPM_STRING_ARRAY_TYPE)
				{
					md5s = (char **) malloc (nr_md5s * sizeof (char **));
					for (i = 0; i < nr_md5s; i++)
						md5s [i] = strdup (((char **) data) [i]);
				}
				else
					caml_failwith ("FILEMD5S tag without string array tyoe");
				break;
			case RPMTAG_PACKAGER: /* string type */
				if ((type == RPM_STRING_TYPE) || (type == RPM_I18NSTRING_TYPE))
					maintainer = strdup (ocamlize_string (my_locale, nr_locales, type, (char *) data));
				else
					caml_failwith ("PACKAGER tag without (i18n) string type");	
				break;
			case RPMTAG_GROUP: /* string type */
				if ((type == RPM_STRING_TYPE) || (type == RPM_I18NSTRING_TYPE))
					section = strdup (ocamlize_string (my_locale, nr_locales, type, (char *) data));
				else
					caml_failwith ("GROUP tag without (i18n) string type");
				break;
			case RPMTAG_ARCHIVESIZE: /* int32 type */
				if (type == RPM_INT32_TYPE)
					archive_size = *((uint_32 *) data);
				else
					caml_failwith ("ARCHIVE_SIZE tag without int32 type");
				break;
			case RPMTAG_SIZE: /* int32 type */
				if (type == RPM_INT32_TYPE)
					installed_size = *((uint_32 *) data);
				else
					caml_failwith ("SIZE tag without int32 type");
				break;
			case RPMTAG_SOURCERPM: /* string type */
#ifdef DEBUG
				fprintf (stderr, "- Source RPM: %s\n", (char *) data);
#endif
				if (type == RPM_STRING_TYPE)
					source_package = strdup ((char *) data);
				else
					caml_failwith ("SOURCERPM tag without string type");
				break;
			case RPMTAG_URL: /* string type */
				if ((type == RPM_STRING_TYPE) || (type == RPM_I18NSTRING_TYPE))
					url = strdup (ocamlize_string (my_locale, nr_locales, type, (char *) data));
				else
					caml_failwith ("URL tag without (i18n) string type");
				break;
			case RPMTAG_SUMMARY: /* string type */
				if ((type == RPM_STRING_TYPE) || (type == RPM_I18NSTRING_TYPE))
					summary = strdup (ocamlize_string (my_locale, nr_locales, type, (char *) data));
				else
					caml_failwith ("SUMMARY tag without (i18n) string type");
				break;
			case RPMTAG_DESCRIPTION: /* string type */
				if ((type == RPM_STRING_TYPE) || (type == RPM_I18NSTRING_TYPE))
					description = strdup (ocamlize_string (my_locale, nr_locales, type, (char *) data));
				else
					caml_failwith ("DESCRIPTION tag without (i18n) string type");
				break;
			case RPMTAG_LICENSE:
#ifdef DEBUG
				fprintf (stderr, "- License: %s\n", (char *) data);
#endif
				if (type == RPM_STRING_TYPE)
					license = strdup ((char *) data);
				else
					caml_failwith ("LICENSE tag without string type");
				break;
			case RPMTAG_DISTRIBUTION:
#ifdef DEBUG
				fprintf (stderr, "- Distribution: %s\n", (char *) data);
#endif
				if (type == RPM_STRING_TYPE)
					distribution = strdup ((char *) data);
				else
					caml_failwith ("DISTRIBUTION tag without string type");
				break;
			case RPMTAG_VENDOR:
				if ((type == RPM_STRING_TYPE) || (type == RPM_I18NSTRING_TYPE))
					vendor = strdup (ocamlize_string (my_locale, nr_locales, type, (char *) data));
				else
					caml_failwith ("VENDOR tag without (i18n) string type");
				break;
			case RPMTAG_OS:
				if (type == RPM_STRING_TYPE)
					os = strdup ((char *) data);
				else
					caml_failwith ("OS tag without string type");
				break;
			/* case RPMTAG_BUILDTIME:
				if (type == RPM_INT32_TYPE)
					date = *((uint_32 *) data);
				else
					caml_failwith ("BUILDTIME tag without int32 type");
				break; */
			case RPMTAG_DISTURL:
				if (type == RPM_STRING_TYPE)
					filename = strdup ((char *) data);
				else
					caml_failwith ("OS tag without string type");
		}
	}
	headerFreeIterator (hi);
	free (locales);
	/* name */
	if (name == NULL)
		caml_failwith ("No NAME tag found");
	else
		Store_field (metadata, METADATA_NAME, caml_copy_string (name));
	/* add all version parts together */
	if (epoch == 0)
		version_string = strdup ("");
	else
		asprintf (&version_string, "%d:", epoch);
	if (version != NULL)
	{
		version_string = (char *) realloc (version_string, strlen (version_string) + strlen (version) + 1);
		strcat (version_string, version);
	}
	if (release != NULL)
	{
		version_string = (char *) realloc (version_string, strlen (version_string) + strlen (release) + 2);
		strcat (version_string, "-");
		strcat (version_string, release);
	}
	Store_field (metadata, METADATA_VERSION, caml_copy_string (version_string));
	free (version); free (release);
	/* dependencies */
	/* v0, v1: pkg_dep_expression (= pkg_disjunction list) */
	v0 = Val_int (0); /* tail */
	for (i = 0; i < nr_d; i++)
	{	
		v1 = caml_alloc (2, 0); /* new head */
		/* v2: pkg_disjunction with one element */
		v2 = caml_alloc (2, 0);
		Store_field (v2, 0, ocamlize_dependency (d_names [i], d_versions [i], d_flags [i]));
		Store_field (v2, 1, Val_int (0));
		Store_field (v1, 0, v2);
		Store_field (v1, 1, v0);
		v0 = v1; /* tail = new head */
	}
	Store_field (metadata, METADATA_DEPENDS, v0);
	/* free stuff */
	for (i = 0; i < nr_d; i++)
	{
		free (d_names [i]);
		free (d_versions [i]);
	}
	free (d_names); free (d_versions); free (d_flags);
	/* conflicts */
	/* v0, v1: pkg_dependency list */
	v0 = Val_int (0); /* tail */
	for (i = 0; i < nr_c; i++)
	{
		v1 = caml_alloc (2, 0);
		Store_field (v1, 0, ocamlize_dependency (c_names [i], c_versions [i], c_flags [i]));
		Store_field (v1, 1, v0);
		v0 = v1; /* tail = new head */
	}
	Store_field (metadata, METADATA_CONFLICTS, v0);
	/* free stuff */
	for (i = 0; i < nr_c; i++)
	{
		free (c_names [i]);
		free (c_versions [i]);
	}
	free (c_names); free (c_versions); free (c_flags);
	v0 = Val_int (0); /* tail */
	for (i = 0; i < nr_o; i++)
	{
		v1 = caml_alloc (2, 0);
		Store_field (v1, 0, ocamlize_dependency (o_names [i], o_versions [i], o_flags [i]));
		Store_field (v1, 1, v0);
		v0 = v1; /* tail = new head */
	}
	Store_field (metadata, METADATA_REPLACES, v0);
	/* free stuff */
	for (i = 0; i < nr_o; i++)
	{
		free (o_names [i]);
		free (o_versions [i]);
	}
	free (o_names); free (o_versions); free (o_flags);
	v0 = Val_int (0); /* tail */
	for (i = 0; i < nr_p; i++)
	{
		v1 = caml_alloc (2, 0);
		Store_field (v1, 0, ocamlize_dependency (p_names [i], p_versions [i], p_flags [i]));
		Store_field (v1, 1, v0);
		v0 = v1; /* tail = new head */
	}
	Store_field (metadata, METADATA_PROVIDES, v0);
	/* free stuff */
	for (i = 0; i < nr_p; i++)
	{
		free (p_names [i]);
		free (p_versions [i]);
	}
	free (p_names); free (p_versions); free (p_flags);
	/* other dependencies are not given by RPM */
	Store_field (metadata, METADATA_RECOMMENDS, Val_int (0));
	Store_field (metadata, METADATA_SUGGESTS, Val_int (0));
	Store_field (metadata, METADATA_PRE_DEPENDS, Val_int (0));
	Store_field (metadata, METADATA_ENHANCES, Val_int (0));
	if (arch == NULL)
	{
#ifdef DEBUG
		fprintf (stderr, "- No architecture given\n");
#endif
		Store_field (metadata, METADATA_ARCHITECTURE, Val_int (1)); /* AnyArchitecture */
	}
	else if (strcmp (arch, "noarch") == 0)
	{
#ifdef DEBUG
		fprintf (stderr, "- All architectures\n");
#endif
		Store_field (metadata, METADATA_ARCHITECTURE, Val_int (0)); /* AllArchitectures */
	}
	else
	{
#ifdef DEBUG
		fprintf (stderr, "- Specific architecture: %s\n", arch);
#endif
		v1 = caml_alloc (1, 0);
		Store_field (v1, 0, caml_copy_string (arch));
		Store_field (metadata, METADATA_ARCHITECTURE, v1);
	}
	free (arch);
	/* files */
	v2 = Val_int (0);
	if ((nr_bn != nr_md5s) && (nr_md5s > 0))
	{
		caml_failwith ("The number of basenames is not equal to the number of MD5 keys");
	}
	for (i = 0; i < nr_bn; i++)
	{
		char *t;
		asprintf (&t, "%s%s", dirnames [dirindexes [i]], basenames [i]);
		v1 = caml_alloc (2, 0);
			v3 = caml_alloc (2, 0);
			Store_field (v3, 0, caml_copy_string (t));
			Store_field (v3, 1, caml_copy_string (nr_md5s == 0 ? "" : md5s [i]));
		Store_field (v1, 0, v3);
		Store_field (v1, 1, v2);
		free (t);
		v2 = v1;
	}
	Store_field (metadata, METADATA_FILES, v2);
	for (i = 0; i < nr_bn; i++)
		free (basenames [i]);
	for (i = 0; i < nr_dn; i++)
		free (dirnames [i]);
	for (i = 0; i < nr_md5s; i++)
		free (md5s [i]);
	free (basenames); free (dirnames); free (dirindexes); free (md5s);
	free (name);
	/* maintainer */
	if (maintainer == NULL)	
		Store_field (metadata, METADATA_MAINTAINER, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_MAINTAINER, caml_copy_string (maintainer));
		free (maintainer);
	}
	/* section */
	if (section == NULL)	
		Store_field (metadata, METADATA_SECTION, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_SECTION, caml_copy_string (section));
		free (section);
	}
	/* size */
	Store_field (metadata, METADATA_SIZE, caml_copy_int32 (archive_size));
	/* installed_size */
	Store_field (metadata, METADATA_INSTALLED_SIZE, caml_copy_int32 (installed_size));
	/* source_package and source_version */
	/* For` the moment, assume <name>-<version>-<release>.src.rpm
	 * with no dashes in <version> and <release> */
	if (source_package == NULL)
	{
		Store_field (metadata, METADATA_SOURCE_PACKAGE, caml_copy_string (""));
		Store_field (metadata, METADATA_SOURCE_VERSION, caml_copy_string (""));
	}
	else
	{
		char *last_dash = strrchr (source_package, '-'), *sep, *dot;
		if (last_dash != NULL)
		{
			dot = strchr (last_dash + 1, '.');
			if (dot != NULL)
				*dot = '\0';
		 	*last_dash = '\0';
			sep = strrchr (source_package, '-');
			if (sep != NULL)
			{
				*last_dash = '-';
				*sep = '\0';
				Store_field (metadata, METADATA_SOURCE_PACKAGE, caml_copy_string (source_package));
				Store_field (metadata, METADATA_SOURCE_VERSION, caml_copy_string (sep + 1));
			}
			else
			{
				Store_field (metadata, METADATA_SOURCE_PACKAGE, caml_copy_string (source_package));
				Store_field (metadata, METADATA_SOURCE_VERSION, caml_copy_string (last_dash + 1));
			}
		}
		else
		{
			Store_field (metadata, METADATA_SOURCE_PACKAGE, caml_copy_string (source_package));
			Store_field (metadata, METADATA_SOURCE_VERSION, caml_copy_string (version_string));
		}
	}
	free (source_package); free (version_string);
	/* url */
	if (url == NULL)	
		Store_field (metadata, METADATA_URL, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_URL, caml_copy_string (url));
		free (url);
	}
	/* summary */
	if (summary == NULL)	
		Store_field (metadata, METADATA_SUMMARY, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_SUMMARY, caml_copy_string (summary));
		free (summary);
	}
	/* description */
	if (description == NULL)	
		Store_field (metadata, METADATA_DESCRIPTION, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_DESCRIPTION, caml_copy_string (description));
		free (description);
	}
	/* license */
	if (license == NULL)	
		Store_field (metadata, METADATA_LICENSE, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_LICENSE, caml_copy_string (license));
		free (license);
	}
	/* distribution */
	if (distribution == NULL)	
		Store_field (metadata, METADATA_DISTRIBUTION, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_DISTRIBUTION, caml_copy_string (distribution));
		free (distribution);
	}
	/* vendor */
	if (vendor == NULL)	
		Store_field (metadata, METADATA_VENDOR, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_VENDOR, caml_copy_string (vendor));
		free (vendor);
	}
	/* essential */
	Store_field (metadata, METADATA_ESSENTIAL, Val_int (0));
	/* os */
	if (os == NULL)	
		Store_field (metadata, METADATA_OS, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_OS, caml_copy_string (os));
		free (os);
	}
	if (filename == NULL)
		Store_field (metadata, METADATA_FILENAME, caml_copy_string (""));
	else
	{
		Store_field (metadata, METADATA_FILENAME, caml_copy_string (filename));
		free (filename);
	}
	/* date */ 
	/* Store_field (metadata, METADATA_DATE, caml_copy_int32 (date)); */
	/* return it all */
	Store_field (metadata, METADATA_BUGS, caml_copy_string (""));
	Store_field (metadata, METADATA_MD5SUM, caml_copy_string (""));
	Store_field (metadata, METADATA_ORIGIN, caml_copy_string (""));
	Store_field (metadata, METADATA_PRIORITY, Val_int (2)); /* standard */
	Store_field (metadata, METADATA_SHA1SUM, caml_copy_string (""));
	Store_field (metadata, METADATA_SHA256SUM, caml_copy_string (""));
	Store_field (metadata, METADATA_TAGS, Val_int (0)); /* [] */
	Store_field (metadata, METADATA_TASKS, Val_int (0)); /* [] */
	CAMLreturn (metadata);
}

value read_rpm_file (value locale, value filename)
{
	FD_t fd;
	Header header;
	struct rpmlead lead;
	CAMLparam2 (locale, filename);
	CAMLlocal1 (retval);
	fd = fdOpen (String_val (filename), O_RDONLY, 0);
	if (!fd)
	 	caml_failwith (strerror (errno));	
	readLead (fd, &lead);
	rpmReadSignature (fd, NULL, lead.signature_type);
	header = headerRead (fd, lead.major >= 3 ? HEADER_MAGIC_YES : HEADER_MAGIC_NO);
	retval = ocamlize_header (header, String_val (locale));	
	fdClose (fd);
	CAMLreturn (retval);
}

value read_hdlist (value locale, value filename)
{
	FD_t fd;
	Header header;
	CAMLparam2 (locale, filename);
	CAMLlocal2 (hd, tl);
	fprintf (stderr, "[RPM] Opening hdlist file: [%s]\n", String_val (filename));
	fd = fdOpen (String_val (filename), O_RDONLY, 0);
	if (!fd)
		caml_failwith (strerror (errno));
	header = headerRead (fd, HEADER_MAGIC_YES);
	tl = Val_int (0);
	while (header != NULL)
	{
		hd = caml_alloc (2, 0);
		Store_field (hd, 0, ocamlize_header (header, String_val (locale)));
		Store_field (hd, 1, tl);
		header = headerRead (fd, HEADER_MAGIC_YES);
		tl = hd;
	}
	fdClose (fd);
	CAMLreturn (tl);
}
